#include <stdio.h>
#include <stdlib.h>
#include <complex.h>

#include "phy/signal.h"
#include "utils/file.h"
#include "phy/frequency_offset.h"
#include "utils/fft.h"
#include "phy/channel.h"
#include "phy/qpsk.h"
#include "transport/fic.h"
#include "coding/frequency_interleaving.h"

void usage(void)
{
  printf("usage: <input file>\n");
  exit(0);
}

int time_sync(file_io *in)
{
  complex ts[T_F + T_NULL];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = get_samples(in, ts, T_F + T_NULL);
  if (c != T_F + T_NULL)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_F + T_NULL; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > T_F)
    min_pos -= T_F;

  put_back_samples(in, &ts[min_pos], T_F + T_NULL - min_pos);

  return min_pos;
}

int fast_time_sync(file_io *in)
{
  complex ts[T_NULL + 200];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = get_samples(in, ts, T_NULL + 200);
  if (c != T_NULL + 200)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_NULL + 200; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > 200)
    min_pos -= 200;

  put_back_samples(in, &ts[min_pos], T_NULL + 200 - min_pos);

  return min_pos;
}

int main(int n, char **v)
{
  char *in_name = NULL;
  file_io *in;
  int i;
  int j;
  int start;
  double fine_freq_offset;
  complex *in_time;
  complex *_cur_freq;
  complex *_prev_freq;
  complex *cur_freq;
  complex *prev_freq;
  complex *channel;
  complex *z_ref;
  fft_data *fft_data;
  int freq_offset_cur_pos;
  complex qpsk[K];
  int p[K*2];
  fic *fic;
  frequency_interleaver *fl;

  for (i = 1; i < n; i++) {
    if (in_name == NULL) { in_name = v[i]; continue; }
    usage();
  }

  if (in_name == NULL) usage();

  fic = new_fic();

  fl = new_frequency_interleaver();

  init_frequency_offset();

  in_time = malloc(sizeof(complex) * T_S);
  _cur_freq = malloc(sizeof(complex) * T_U);
  _prev_freq = malloc(sizeof(complex) * T_U);
  channel = malloc(sizeof(complex) * T_U);
  z_ref = malloc(sizeof(complex) * T_U);
  if (in_time == NULL || _cur_freq == NULL || _prev_freq == NULL
      || channel == NULL || z_ref == NULL) abort();

  cur_freq = _cur_freq;
  prev_freq = _prev_freq;

  fft_data = init_fft(T_U);

  compute_phase_ref_full(fft_data, z_ref);

  in = new_input_file(in_name);
  if (in == NULL) exit(1);

  start = time_sync(in);
  if (start == -1) goto over;

next_frame:
  printf("#start %d\n", start);

  if (get_samples(in, NULL, T_NULL) != T_NULL) goto over;

  /* symbol 1 */
  if (get_samples(in, in_time, T_S) != T_S) goto over;
  fine_freq_offset = frequency_offset_fine(in_time);
  printf("# %g ", fine_freq_offset);
  freq_offset_cur_pos = remove_frequency_offset(0, 
                            in_time, fine_freq_offset);
  fine_freq_offset = frequency_offset_fine(in_time);
  printf("%g\n", fine_freq_offset);

  fft(fft_data, &in_time[Delta], prev_freq);

  //channel_estimate(z_ref, prev_freq, channel);
  //channel_compensate(prev_freq, channel, prev_freq);

  for (i = 2; i < 77; i++) {
    if (get_samples(in, in_time, T_S) != T_S) goto over;
    fine_freq_offset = frequency_offset_fine(in_time);
    printf("#%g ", fine_freq_offset);
    freq_offset_cur_pos = remove_frequency_offset(freq_offset_cur_pos,
                              in_time, fine_freq_offset);
    fine_freq_offset = frequency_offset_fine(in_time);
    printf("%g\n", fine_freq_offset);

    fft(fft_data, &in_time[Delta], cur_freq);

    //channel_compensate(cur_freq, channel, cur_freq);

    /* frequency de-interleaving done here (using k2n) */
    /* positive frequencies [1..768] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j+1)]] = cur_freq[j+1] * conj(prev_freq[j+1]);
    /* negative frequencies [-768..-1] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j-K/2)]] = cur_freq[j+T_U-K/2] * conj(prev_freq[j+T_U-K/2]);

    symbol_unmap(p, qpsk);

    /* symbols 2, 3 and 4 are for fic */
    if (i < 5) {
      for (j = 0; j < 4; j++) {
        /* FIB = 256 bits, 30 bytes of data, 2 bytes of crc
         * 4 FIBs per symbol, 12 FIBs for the 3 symbols of the FIC
         * 3 FIBs make for 1 FIC and are used for energy dispersal (768 bits)
         * FIC (768 bits) is convolution encoded, giving 3072+24 bits
         * 3072+24 bits are punctured, giving 2304 bits
         * puncturing is:
         *     21 blocks of 128 bits are punctured by PI=16 (rate 8/24)
         *     (we remove 8 bits over 32)
         *     3 blocks of 128 bits are punctured by PI=15 (rate 9/24)
         *     (we remove 9 bits over 32)
         *     last 24 bits are punctured by 1100 1100 1100 1100 .. 1100
         *     (we remove 12 bits over 24)
         * each symbol gets 3072 bits, so:
         * 1st symbol: fic 0 (2304 bits) + fic 1 (768 bits)
         * 2nd symbol: fic 1 (1536 bits) + fic 2 (1536 bits)
         * 3rd symbol: fic 2 (768 bits)  + fic 3 (2304 bits)
         * we send to the fic unit blocks of 768 bits
         * when the fic unit has enough bits it does the decoding
         */
        receive_fic_bits(fic, &p[768*j]);
      }
    }

#if 0
{
  int i;

static int tt = 0;
if (tt == 2500)
  for (i = 0; i < K; i++) printf("%g %g\n", creal(qpsk[i]), cimag(qpsk[i]));
if (tt == 2500)
exit(0);
tt++;
}
#endif
#define  SWAP(a, b) do { complex *x = a; a = b; b = x; } while (0)
    SWAP(cur_freq, prev_freq);
  }

  /* put back last 100 samples of last symbol, for fast time sync */
  put_back_samples(in, &in_time[T_S-100], 100);
  start = fast_time_sync(in);
  if (start == -1) goto over;

  goto next_frame;

over:
  close_input(in);

  return 0;
}
